#!/usr/bin/ruby
# vim:fileencoding=utf-8
#
# Copyright (c) 2013 ModCloth, Inc.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

require 'fileutils'
require 'json'
require 'minitest/spec'
require 'net/http'
require 'time'
require 'uri'

begin
  require_relative 'mtbb/version'
rescue LoadError
end

module Mtbb
  VERSION = '0.1.2' unless defined?(VERSION)

  YELLOW = "\033\[33m"
  RESET = "\033\[0m"
  GREEN = "\033\[32m"
  RED = "\033\[31m"
  MAGENTA = "\033\[35m"
  BRIGHT_GREEN = "\033\[32;1m"
  BRIGHT_RED = "\033\[31;1m"
  BRIGHT_MAGENTA = "\033\[35;1m"
  BRIGHT_YELLOW = "\033\[33;1m"

  SERVERS = {}

  def register(server_name, server_config)
    SERVERS[server_name.to_sym] = Mtbb::ServerRunner.new(server_config)
  end

  def server(server_name)
    SERVERS[server_name.to_sym]
  end

  module_function :register, :server

  class TestRunner
    def self.main(argv = [].freeze)
      mtbb_test_glob = ENV['MTBB_TEST_GLOB'] || "#{Dir.pwd}/test/**/*_test.rb"
      Dir.glob(mtbb_test_glob).each do |f|
        require f
      end

      unless argv.include?('-h') || argv.include?('--help')
        ServerControl.start_servers
        Mtbb.announce! 'Started servers'
        at_exit do
          ServerControl.stop_servers
          Mtbb.announce! 'Stopped servers'
        end

        MiniTest::Unit.output = MiniTestReporter.new
      end

      exit_code = MiniTest::Unit.new.run(argv) || 1

      if exit_code == 0
        $stderr.puts BRIGHT_GREEN
        $stderr.puts <<-EOF.gsub(/^ {8}/, '')
          ✓✓      ✓✓ ✓✓✓✓ ✓✓    ✓✓
          ✓✓  ✓✓  ✓✓  ✓✓  ✓✓✓   ✓✓
          ✓✓  ✓✓  ✓✓  ✓✓  ✓✓✓✓  ✓✓
          ✓✓  ✓✓  ✓✓  ✓✓  ✓✓ ✓✓ ✓✓
          ✓✓  ✓✓  ✓✓  ✓✓  ✓✓  ✓✓✓✓
          ✓✓  ✓✓  ✓✓  ✓✓  ✓✓   ✓✓✓
           ✓✓✓  ✓✓✓  ✓✓✓✓ ✓✓    ✓✓
        EOF
        $stderr.puts RESET
      else
        $stderr.puts BRIGHT_RED
        $stderr.puts <<-EOF.gsub(/^ {8}/, '')
          ✘✘✘✘✘✘✘✘    ✘✘✘    ✘✘✘✘ ✘✘
          ✘✘         ✘✘ ✘✘    ✘✘  ✘✘
          ✘✘        ✘✘   ✘✘   ✘✘  ✘✘
          ✘✘✘✘✘✘   ✘✘     ✘✘  ✘✘  ✘✘
          ✘✘       ✘✘✘✘✘✘✘✘✘  ✘✘  ✘✘
          ✘✘       ✘✘     ✘✘  ✘✘  ✘✘
          ✘✘       ✘✘     ✘✘ ✘✘✘✘ ✘✘✘✘✘✘✘✘
        EOF
        $stderr.puts RESET
        Mtbb::SERVERS.each do |_, server|
          Mtbb.announce! "Dumping stdout for #{server.description}:"
          server.dump(:stdout)
          Mtbb.announce! "Dumping stderr for #{server.description}:"
          server.dump(:stderr)
        end
      end

      if ENV['HOLD_ON_A_SEC']
        print 'Holding on a sec... [Enter] '
        gets
      end
      exit exit_code
    end
  end

  def announce!(something)
    $stderr.puts "#{BRIGHT_GREEN}mtbb#{RESET}: #{BRIGHT_YELLOW}#{something}#{RESET}"
  end

  module_function :announce!

  class ServerRunner
    attr_reader :executable, :server_name, :argv, :port, :working_dir, :start_time
    attr_reader :server_pid, :stdout_file, :stderr_file, :server_exit_code
    attr_reader :options

    def initialize(options = {})
      @options = options
      @executable = options.fetch(:executable)
      options.delete(:executable)
      @argv = options.delete(:argv) || []

      @server_name = options.delete(:server_name) || "mtbb-server-#{Time.now.utc.to_i}"
      @start_time = options.delete(:start) || Time.now.utc
      @port = options.delete(:port)
      @working_dir = options.delete(:working_dir) || default_working_dir
      @stdout_file = options.delete(:stdout_file) || default_stdout_file
      @stderr_file = options.delete(:stderr_file) || default_stderr_file

      if !File.exist?(executable)
        raise "Can't locate `#{File.basename(executable).inspect}` executable! " <<
              "(it's not here: #{executable.inspect})"
      end
    end

    def start
      FileUtils.mkdir_p(working_dir)
      process_command = command
      spawn_args = process_command + [
        out: [stdout_file, 'w'], err: [stderr_file, 'w']
      ]
      Mtbb.announce! "Starting #{description}"
      Mtbb.announce! "  ---> #{process_command.join(' ')}"
      Mtbb.announce! "  (#{config_details.inspect})"
      @server_pid = Process.spawn(*spawn_args)
      server_pid
    end

    def stop
      return if !server_pid
      Mtbb.announce! "Stopping #{description} (PID=#{server_pid})"
      Process.kill(:TERM, server_pid)
      begin
        wait_for_server_process(Integer(ENV['MTBB_PROCESS_EXIT_TIMEOUT'] || 3))
      rescue TimeoutError
      end
      Process.kill(:KILL, server_pid)
      _, status = Process.waitpid2(server_pid)
      @server_exit_code = status.exitstatus || status.termsig
    rescue Errno::ECHILD, Errno::ESRCH
      true
    end

    def description
      "#{server_name} on port #{port}"
    end

    def config_details
      {
        argv: argv,
        working_dir: working_dir,
        start_time: start_time,
        stdout_file: stdout_file,
        stderr_file: stderr_file,
      }
    end

    def command
      cmd = [executable] + argv
      options.each do |k,v|
        if v.nil?
          cmd << k
          next
        end
        cmd << k << v.inspect
      end
      cmd
    end

    def dump(which = :stdout)
      $stderr.puts File.read(send(:"#{which.to_s}_file"))
    end

    private

    def wait_for_server_process(timeout)
      toolate = Time.now + timeout
      unless (done = server_exited?) || Time.now > toolate
        sleep Float(ENV['MTBB_PROCESS_EXIT_POLL_INTERVAL'] || 0.1)
      end

      unless done
        raise TimeoutError
      end
    end

    # Largely inspired by ChildProcess::Unix::Process#exited?
    def server_exited?
      return true if server_exit_code

      pid, status = Process.waitpid2(server_pid, Process::WNOHANG)
      pid = pid == 0 ? nil : pid

      if pid
        @server_exit_code = status.exitstatus || status.termsig
      end

      !!pid
    end

    def default_working_dir
      File.join(Dir.pwd, '.mtbb-artifacts')
    end

    def default_stdout_file
      File.join(working_dir, "#{server_name}.out")
    end

    def default_stderr_file
      File.join(working_dir, "#{server_name}.err")
    end
  end

  class MiniTestReporter
    def puts(*args)
      args.each { |arg| Mtbb.announce! arg }
    end

    alias print puts
  end

  module NetThings
    def http(method_name, path, options = {})
      options[:path] ||= path
      send("#{method_name.downcase}_request", options)
    end

    def post_request(options = {})
      request = Net::HTTP::Post.new(options[:path] || '/')
      request.content_type = options[:content_type] || 'application/x-www-form-urlencoded'
      request.body = options.fetch(:body)
      perform_request(request, options.fetch(:port))
    end

    def get_request(options = {})
      perform_request(
        Net::HTTP::Get.new(options[:path] || '/'),
        options.fetch(:port)
      )
    end

    def delete_request(options = {})
      perform_request(
        Net::HTTP::Delete.new(options[:path] || '/'),
        options.fetch(:port)
      )
    end

    private

    def perform_request(request, port)
      Net::HTTP.start(ENV['MTBB_SERVER_HOST'] || 'localhost', port) do |http|
        http.open_timeout = Integer(ENV['MTBB_HTTP_OPEN_TIMEOUT'] || 1)
        http.read_timeout = Integer(ENV['MTBB_HTTP_READ_TIMEOUT'] || 3)
        response = http.request(request)
        response
      end
    end
  end

  class ServerControl
    def self.stop_servers
      Mtbb::SERVERS.each do |_,server|
        server.stop
      end
    end

    def self.start_servers
      Dir.chdir(File.expand_path('../', __FILE__)) do
        FileUtils.mkdir_p('./.mtbb-artifacts')
        Mtbb::SERVERS.each do |_,runner|
          runner.start
          wait_for_server(runner.port)
        end
      end
    end

    def self.wait_for_server(http_port)
      maxloops = 10
      curloop = 0
      begin
        TCPSocket.new(ENV['MTBB_SERVER_HOST'] || 'localhost', http_port).close
      rescue
        curloop += 1
        if curloop < maxloops
          sleep 0.5
          retry
        end
      end
    end
  end
end

if __FILE__ == $0
  $:.unshift ENV['PWD']
  Mtbb::TestRunner.main(ARGV)
end
